import contextlib
import logging
import os
import pathlib
import shutil
from typing import Iterable, Tuple


logger = logging.getLogger('caterpillar')
_fmt = logging.Formatter(fmt='[%(levelname)s] %(message)s')
_sh = logging.StreamHandler()
_sh.setFormatter(_fmt)
logger.addHandler(_sh)
logger.setLevel(logging.WARNING)


def increase_logging_verbosity(num_levels):
    target_level = logger.level - num_levels * 10
    target_level = min(max(target_level, logging.DEBUG), logging.CRITICAL)
    logger.setLevel(target_level)


def should_log_error():
    return logger.isEnabledFor(logging.ERROR)


def should_log_warning():
    return logger.isEnabledFor(logging.WARNING)


def should_log_info():
    return logger.isEnabledFor(logging.INFO)


def should_log_debug():
    return logger.isEnabledFor(logging.DEBUG)


# Returns the qualified name of an exeception.
def excname(value):
    etype = type(value)
    if etype.__module__ == 'builtins':
        return etype.__name__
    else:
        return '%s.%s' % (etype.__module__, etype.__name__)


# Resolve a pathlib.Path that may not exist yet. Assumes that the parent
# of the path already exists.
#
# This is to workaround pathlib.Path.resolve() behavior on Windows: if
# the relative path does not already exist, the return value is not an
# absolute path.
def abspath(path: pathlib.Path) -> pathlib.Path:
    return path.parent.resolve().joinpath(path.name)


@contextlib.contextmanager
def chdir(directory):
    cwd = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(cwd)


# Monkey patch shutil.get_terminal_size to fix full-width progress bar
# overflow problem on Windows consoles.
def monkeypatch_get_terminal_size():
    # Only monkey patch on NT, and only monkey patch once.
    if os.name != 'nt' or hasattr(shutil, 'original_get_terminal_size'):
        return
    shutil.original_get_terminal_size = shutil.get_terminal_size

    def replacement(fallback=None):
        columns, lines = shutil.original_get_terminal_size(fallback)
        # One fewer column so that full-width progress bar doesn't flow
        # onto the next line.
        return os.terminal_size((columns - 1, lines))

    shutil.get_terminal_size = replacement


# A stub class with support for random attribute access. All attributes
# not previously set are regarded as a stub method that takes any
# positional and keyword arguments and returns None.
class Stub(object):

    def __getattr__(self, name):
        def stub(*_, **__):
            return

        return stub

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)


# A context manager factory that yields a Stub object and does no more.
@contextlib.contextmanager
def stub_context_manager(*_, **__):
    yield Stub()


# A bare minimum M3U8 generator (HLSv3).
#
# segments is an iterable of tuples (url, duration).
#
# Note that the only required media playlist tag is
# EXT-X-TARGETDURATION, and the only required media segment tag is
# EXTINF. Additionally, we use 3 as EXT-X-VERSION for floating-point
# EXTINF duration values.[1]
#
# [1] https://tools.ietf.org/html/rfc8216#section-7
def generate_m3u8(target_duration: int, segments: Iterable[Tuple[str, float]]):
    lines = []
    lines.append('#EXTM3U')
    lines.append('#EXT-X-VERSION:3')
    lines.append(f'#EXT-X-TARGETDURATION:{target_duration}')
    for url, duration in segments:
        lines.append(f'#EXTINF:{duration},')
        lines.append(url)
    lines.append('#EXT-X-ENDLIST')
    lines.append('')
    return '\n'.join(lines)



import multiprocessing
import os
import pathlib
import signal
import time
import urllib.parse
from typing import Tuple

import click
import m3u8
import requests

from .utils import (
    excname,
    generate_m3u8,
    logger,
    monkeypatch_get_terminal_size,
    should_log_warning,
    stub_context_manager,
)


CHUNK_SIZE = 65536  # Download chunk size (64K)
REQUESTS_TIMEOUT = 5  # Both connect timeout and read timeout
MAX_RETRY_INTERVAL = 30  # Upper bound on exponential backoff

# For proper progress bar rendering on Windows consoles.
monkeypatch_get_terminal_size()


# Returns a bool indicating success (True) or failure (False).
def resumable_download(url: str, file: pathlib.Path) -> bool:
    existing_bytes = file.stat().st_size if file.is_file() else 0
    try:
        logger.debug(f'GET {url}')
        r = requests.get(url, headers={'Range': f'bytes={existing_bytes}-'},
                         stream=True, timeout=REQUESTS_TIMEOUT)
        if r.status_code not in {200, 206}:
            logger.error(f'GET {url}: HTTP {r.status_code}')
            return False
        with open(file, 'ab') as fp:
            for chunk in r.iter_content(chunk_size=CHUNK_SIZE):
                if chunk:
                    fp.write(chunk)
        return True
    except Exception as e:
        logger.warning(f'GET {url}: {excname(e)}: {e}')
        return False


# Returns a bool indicating success (True) or failure (False).
def resumable_download_with_retries(url: str, file: pathlib.Path,
                                    max_retries: int = 2) -> bool:
    incomplete_file = file.with_suffix(file.suffix + '.incomplete')

    # If the file, without the .incomplete suffix, is already present,
    # assume it has been downloaded.
    if file.exists():
        return True

    retries = 0
    while True:
        if resumable_download(url, incomplete_file):
            os.replace(incomplete_file, file)
            return True

        if retries >= max_retries:
            logger.error(f'GET {url}: failed after {max_retries} retries')
            return False

        retries += 1
        wait_time = min(2 ** retries, MAX_RETRY_INTERVAL)
        logger.warning(f'GET {url}: retrying after {wait_time} seconds...')
        time.sleep(wait_time)


# Returns a bool indicating success (True) or failure (False).
def download_m3u8_file(m3u8_url: str, file: pathlib.Path) -> bool:
    logger.info(f'downloading {m3u8_url} to {file} ...')
    return resumable_download_with_retries(m3u8_url, file)


# Returns a bool indicating success (True) or failure (False).
def download_segment(url: str, index: int, directory: pathlib.Path,
                     max_retries: int = 2) -> bool:
    return resumable_download_with_retries(url, directory / f'{index}.ts', max_retries=max_retries)


# download_segment wrapper that takes all arguments as a single tuple,
# so that we can use it with multiprocessing.pool.Pool.map and company.
def _download_segment_mappable(args: Tuple[str, int, pathlib.Path]) -> bool:
    return download_segment(*args)


def _init_worker():
    # Ignore SIGINT in worker processes to disable traceback from
    # each worker on keyboard interrupt.
    signal.signal(signal.SIGINT, signal.SIG_IGN)


# Download all segments in remote_m3u8_file (downloaded from
# remote_m3u8_url), and generates a local playlist in local_m3u8_file
# with local segment filenames (0.ts, 1.ts, 2.ts, etc.).
#
# jobs indicates the maximum number of parallel downloads. Default is
# twice os.cpu_count().
#
# Returns a bool indicating success (True) or failure (False).
def download_m3u8_segments(remote_m3u8_url: str,
                           remote_m3u8_file: pathlib.Path,
                           local_m3u8_file: pathlib.Path,
                           jobs: int = None) -> bool:
    if jobs is None:
        jobs = os.cpu_count() * 2

    try:
        remote_m3u8_obj = m3u8.load(remote_m3u8_file.as_posix())
    except Exception as e:
        logger.error(f'failed to parse {remote_m3u8_file}: {excname(e)}: {e}')
        return False

    target_duration = remote_m3u8_obj.target_duration
    local_segments = []
    download_args = []
    for index, segment in enumerate(remote_m3u8_obj.segments):
        url = urllib.parse.urljoin(remote_m3u8_url, segment.uri)
        download_args.append((url, index, local_m3u8_file.parent))
        local_segments.append((f'{index}.ts', segment.duration))

    with open(local_m3u8_file, 'w') as fp:
        fp.write(generate_m3u8(target_duration, local_segments))
    logger.info(f'generated {local_m3u8_file}')

    with multiprocessing.Pool(jobs, _init_worker) as pool:
        total = len(download_args)
        num_success = 0
        num_failure = 0
        logger.info(f'downloading {total} segments...')
        progress_bar_generator = (click.progressbar if should_log_warning() else
                                  stub_context_manager)
        progress_bar_props = dict(
            width=0,  # Full width
            bar_template='[%(bar)s] %(info)s',
            show_pos=True,
            length=total,
        )
        with progress_bar_generator(**progress_bar_props) as bar:
            for success in pool.imap_unordered(_download_segment_mappable, download_args):
                if success:
                    num_success += 1
                else:
                    num_failure += 1
                logger.debug(f'progress: {num_success}/{num_failure}/{total}')
                bar.update(1)

        if num_failure > 0:
            logger.error(f'failed to download {num_failure} segments')
            return False
        else:
            logger.info(f'finished downloading all {total} segments')
            return True



#!/usr/bin/env python3

import pathlib

import setuptools


HERE = pathlib.Path(__file__).resolve().parent
with open(HERE / 'caterpillar' / 'version.py') as fp:
    exec(fp.read())

setuptools.setup(
    name='caterpillar-hls',
    version=__version__,
    description='Hardened HLS merger',
    long_description='See https://github.com/zmwangx/caterpillar#readme.',
    url='https://github.com/zmwangx/caterpillar',
    author='Zhiming Wang',
    author_email='zmwangx@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Conversion',
    ],
    keywords='HLS streaming m3u8 concatenate merge',
    packages=['caterpillar'],
    install_requires=['click', 'm3u8', 'requests'],
    entry_points={
        'console_scripts': [
            'caterpillar=caterpillar.caterpillar:main',
        ],
    },
)



import os
import pathlib
import re
import shutil
import subprocess
import sys
import time
from typing import Tuple

import m3u8

from .utils import (
    abspath,
    chdir,
    generate_m3u8,
    logger,
    should_log_info,
)


# If ignore_errors is True, blast through non-monotonous DTS errors
# without looking back. We use this after on a splitted playlist deemed
# all good, since for some mysterious reason, probably due to artifacts
# in some hopelessly bad segments, it seems possible that a
# non-monotonous DTS error would manifest only on the second pass, after
# the playlist is splitted. An example:
#
#   http://live.us.sinaimg.cn/000XDYqUjx07gRaRHSCz070d010002TZ0k01.m3u8
#
# On the first pass of merging 4.m3u8, we only get the non-monotonous
# DTS error at 13.ts, but after splitting, on the second pass, we also
# get a non-monotonous error from 12.ts, following a "missing picture in
# access unit with size 6" error.
#
# Returns None if the merge succeeds, or the basename of the first bad
# segment if non-monotonous DTS is detected.
def attempt_merge(m3u8_file: pathlib.Path, output: pathlib.Path,
                  ignore_errors: bool = False) -> str:
    logger.info(f'attempting to merge {m3u8_file} into {output}')

    m3u8_obj = m3u8.load(m3u8_file.as_posix())
    if len(m3u8_obj.segments) == 1:
        # Only one segment, cannot further subdivide, so ignore whatever
        # problems there may be.
        logger.info(f'only one segment in playlist; ignoring errors and warnings')
        ignore_errors = True

    regular_pattern = re.compile(r"Opening '(?P<path>.*\.ts)' for reading")
    error_pattern = re.compile('Non-monotonous DTS in output stream')
    command = ['ffmpeg', '-hide_banner', '-loglevel', 'info',
               '-f', 'hls', '-i', m3u8_file.as_posix(), '-c', 'copy', '-y', output.as_posix()]
    p = subprocess.Popen(command, stdin=subprocess.DEVNULL, stderr=subprocess.PIPE,
                         universal_newlines=True, bufsize=1,
                         encoding='utf-8', errors='backslashreplace')
    last_read_segment = None
    for line in p.stderr:
        m = regular_pattern.search(line)
        # Suppress the line if logging level is below INFO and it's just
        # a boring "Opening '...' for reading" message.
        if not m or should_log_info():
            sys.stderr.write(line)
            sys.stderr.flush()
        if m:
            last_read_segment = os.path.basename(m['path'])
            continue
        if ignore_errors:
            continue
        if error_pattern.search(line):
            assert last_read_segment
            logger.warning(f'DTS jump detected in {last_read_segment}')
            if last_read_segment == m3u8_obj.segments[0].uri:
                logger.warning(f'{last_read_segment} is the first segment in playlist; '
                               f'splitting at the next segment')
                split_point = m3u8_obj.segments[1].uri
            else:
                split_point = last_read_segment

            p.stderr.close()
            p.terminate()
            # Deal with Windows process and file ownership idiosyncrasies.
            # On *ix this is immediate.
            while True:
                try:
                    output.unlink()
                except PermissionError:
                    # On Windows, the ffmpeg subprocess is not yet
                    # cleaned up and is still clinging to this file;
                    # wait until it is released, or the next subprocess
                    # may not even be able to successfully overwrite
                    # this file.
                    time.sleep(0.1)
                except FileNotFoundError:
                    break
                else:
                    break

            return split_point
    returncode = p.wait()
    if returncode != 0:
        logger.error(f'ffmpeg failed with exit status {returncode}')
        raise RuntimeError('unknown error occurred during merging')
    else:
        return None


# Split the source m3u8 file into two destination m3u8 files, at
# split_point, which is the URL of a segment. split_point belongs to the
# second file after splitting.
#
# It's safe to overwrite the source file with one of the destinations.
def split_m3u8(source: pathlib.Path, destinations: Tuple[pathlib.Path, pathlib.Path],
               split_point: str) -> None:
    logger.info(f'splitting {source} at {split_point}')
    m3u8_obj = m3u8.load(source.as_posix())
    target_duration = m3u8_obj.target_duration
    part1_segments = []
    part2_segments = []
    reached_split_point = False
    for segment in m3u8_obj.segments:
        if segment.uri == split_point:
            reached_split_point = True
        tup = (segment.uri, segment.duration)
        if reached_split_point:
            part2_segments.append(tup)
        else:
            part1_segments.append(tup)
    dest1, dest2 = destinations
    with open(dest1, 'w') as fp:
        fp.write(generate_m3u8(target_duration, part1_segments))
    logger.info(f'wrote {dest1}')
    with open(dest2, 'w') as fp:
        fp.write(generate_m3u8(target_duration, part2_segments))
    logger.info(f'wrote {dest2}')


# concat_method is either 'concat_demuxer'[1] or 'concat_protocol'[2].
# Sometimes one works better than other, but there's no clear winner in all
# cases.
#
# m3u8_file should not be named '1.m3u8'; in fact, avoid naming it
# '<number>.m3u8', or it may be overwritten in the process.
#
# [1] https://ffmpeg.org/ffmpeg-all.html#concat-1
# [2] https://ffmpeg.org/ffmpeg-all.html#concat-2
def incremental_merge(m3u8_file: pathlib.Path, output: pathlib.Path,
                      concat_method: str = 'concat_demuxer'):
    # Resolve output so that we don't write to a different relative path
    # later when we run FFmpeg from a different pwd.
    output = abspath(output)
    directory = m3u8_file.parent
    playlist_index = 1
    playlist = directory / f'{playlist_index}.m3u8'
    shutil.copyfile(m3u8_file, playlist)

    intermediate_dir = directory / 'intermediate'
    intermediate_dir.mkdir(exist_ok=True)

    while True:
        merge_dest = intermediate_dir / f'{playlist_index}.ts'
        split_point = attempt_merge(playlist, merge_dest)
        if not split_point:
            break
        playlist_index += 1
        next_playlist = directory / f'{playlist_index}.m3u8'
        split_m3u8(playlist, (playlist, next_playlist), split_point)
        attempt_merge(playlist, merge_dest, ignore_errors=True)
        playlist = next_playlist

    with chdir(intermediate_dir):
        if concat_method == 'concat_demuxer':
            with open('concat.txt', 'w') as fp:
                for index in range(1, playlist_index + 1):
                    print(f'file {index}.ts', file=fp)

            command = ['ffmpeg', '-hide_banner', '-loglevel', 'info',
                       '-f', 'concat', '-i', 'concat.txt',
                       '-c', 'copy', '-movflags', 'faststart', '-y', output.as_posix()]
        elif concat_method == 'concat_protocol':
            ffmpeg_input = 'concat:' + '|'.join(f'{i}.ts' for i in range(1, playlist_index + 1))
            command = ['ffmpeg', '-hide_banner', '-loglevel', 'info', '-i', ffmpeg_input,
                       '-c', 'copy', '-movflags', 'faststart', '-y', output.as_posix()]
        else:
            raise NotImplementedError(f"unrecognized concat method '{concat_method}'")

        try:
            logger.info('merging intermediate products...')
            subprocess.run(command, stdin=subprocess.DEVNULL)
        except subprocess.CalledProcessError as e:
            logger.error(f'ffmpeg failed with exit status {e.returncode}')
            raise RuntimeError('unknown error occurred during merging')
        else:
            logger.info(f'merged into {output}')



#!/usr/bin/env python3

import argparse
import pathlib
import shutil
import sys
import urllib.parse

from . import download, merge
from .utils import excname, logger, increase_logging_verbosity
from .version import __version__


def main():
    parser = argparse.ArgumentParser()
    add = parser.add_argument
    add('m3u8_url',
        help='the VOD URL')
    add('output', nargs='?', type=pathlib.Path, default=None,
        help='''path to the final output file (default is a .ts file in the
        current directory with the basename of the VOD URL)''')
    add('-f', '--force', action='store_true',
        help='overwrite the output file if it already exists')
    add('-j', '--jobs', type=int, default=None,
        help='''maximum number of concurrent downloads (default is twice
        the number of CPU cores, including virtual cores)''')
    add('-k', '--keep', action='store_true',
        help='keep intermediate files even after a successful merge')
    add('-m', '--concat-method',
        choices=['concat_demuxer', 'concat_protocol', '0', '1'],
        default='concat_demuxer',
        help='''method for concatenating intermediate files (default is
        'concat_demuxer'); see https://github.com/zmwangx/caterpillar/#notes
        for details''')
    add('--wipe', action='store_true',
        help='wipe all downloaded files (if any) and start over')
    add('-v', '--verbose', action='count', default=0,
        help='increase logging verbosity (can be specified multiple times)')
    add('-q', '--quiet', action='count', default=0,
        help='decrease logging verbosity (can be specified multiple times)')
    add('-V', '--version', action='version', version=__version__)
    args = parser.parse_args()

    increase_logging_verbosity(args.verbose - args.quiet)

    m3u8_url = args.m3u8_url
    output = args.output

    if output is None:
        stem = pathlib.Path(urllib.parse.urlsplit(m3u8_url).path).stem
        if not stem or stem.startswith('.'):
            logger.critical(f'cannot auto-determine an output file from {m3u8_url}')
            return 1
        output = pathlib.Path(f'{stem}.mp4')
        logger.info(f'output not specified; using {output}')

    if not output.parent.exists():
        logger.critical(f'{output.parent} does not exist')
        return 1

    if not output.suffix:
        logger.critical(f'output must have a suffix, e.g., .mp4')
        return 1

    if output.exists() and not args.force:
        logger.critical(f'{output} already exists; specify --force to overwrite it')
        return 1

    if not output.exists():
        # Make sure output (especially if it's auto-deduced from URL, which
        # might contain reserved characters on Windows) is a valid path and is
        # writable.
        try:
            with open(output, 'wb'):
                pass
            output.unlink()
        except OSError:
            logger.critical(f'{output} is not a valid path or is not writable')
            return 1

    if args.jobs is not None and args.jobs <= 0:
        logger.critical('jobs must be positive')
        return 1

    if args.concat_method == '0':
        args.concat_method = 'concat_demuxer'
    elif args.concat_method == '1':
        args.concat_method = 'concat_protocol'

    remote_m3u8_url = m3u8_url
    working_directory = output.with_suffix('')
    if args.wipe and working_directory.exists():
        logger.info(f'wiping {working_directory}')
        try:
            if working_directory.is_file():
                working_directory.unlink()
            else:
                shutil.rmtree(working_directory)
        except OSError as e:
            logger.critical(f'failed to wipe {working_directory}: {excname(e)}: {e}')
            return 1
    working_directory.mkdir(exist_ok=True)
    remote_m3u8_file = working_directory / 'remote.m3u8'
    local_m3u8_file = working_directory / 'local.m3u8'
    try:
        if not download.download_m3u8_file(remote_m3u8_url, remote_m3u8_file):
            logger.critical(f'failed to download {remote_m3u8_url}')
            return 1
        logger.info(f'downloaded {remote_m3u8_file}')
        if not download.download_m3u8_segments(remote_m3u8_url,
                                               remote_m3u8_file,
                                               local_m3u8_file,
                                               jobs=args.jobs):
            logger.critical('failed to download some segments')
            return 1
        merge.incremental_merge(local_m3u8_file, output,
                                concat_method=args.concat_method)
        if not args.keep:
            shutil.rmtree(working_directory)
    except RuntimeError as e:
        logger.critical(e)
        return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())
